--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

include("keysym.h.lua")


function widget:GetInfo()
  return {
    name      = "Auto group",
    desc      = "Assigns new units from factory to predefined groups by unit types. Ctrl+alt+0-9 setups auto group from currently selected units. Ctrl+al+\\ deletes any autogrouping for given unit type.",
    author    = "Licho",
    date      = "Mar 23, 2007",
    license   = "GNU GPL, v2 or later",
    layer     = 0,
    enabled   = true  --  loaded by default?
  }
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local finiGroup = {}
local unit2group = {}
local myTeam

-- speedups
local SetUnitGroup = Spring.SetUnitGroup
local GetSelectedUnits = Spring.GetSelectedUnits
local GetUnitDefID = Spring.GetUnitDefID
local Echo = Spring.Echo

function widget:Initialize() 
	local _, _, spec, team = Spring.GetPlayerInfo(Spring.GetMyPlayerID())
	if spec then
		widgetHandler:RemoveWidget()
		return false
	end
  myTeam = team
end


function widget:UnitFinished(unitID, unitDefID, unitTeam)
	if (unitTeam == myTeam and unitID ~= nil) then
    finiGroup[unitID] = 1
  end
end


function widget:UnitDestroyed(unitID, unitDefID, teamID)
  finiGroup[unitID] = nil
end

function widget:UnitGiven(unitID, unitDefID, newTeamID, teamID)
  finiGroup[unitID] = nil
end


function widget:UnitIdle(unitID, unitDefID, unitTeam) 
  if (unitTeam == myTeam and finiGroup[unitID]~=nil) then
    local gr = unit2group[unitDefID]
    if gr ~= nil then SetUnitGroup(unitID, gr) end
    finiGroup[unitID] = nil
  end
end


function widget:KeyPress(key, modifier, isRepeat)
	if (modifier.ctrl and modifier.alt) then
		local gr
		if (key == KEYSYMS.N_0) then gr = 0 end
		if (key == KEYSYMS.N_1) then gr = 1 end
		if (key == KEYSYMS.N_2) then gr = 2 end 
		if (key == KEYSYMS.N_3) then gr = 3 end
		if (key == KEYSYMS.N_4) then gr = 4 end
		if (key == KEYSYMS.N_5) then gr = 5 end
		if (key == KEYSYMS.N_6) then gr = 6 end
		if (key == KEYSYMS.N_7) then gr = 7 end
		if (key == KEYSYMS.N_8) then gr = 8 end
		if (key == KEYSYMS.N_9) then gr = 9 end
 		if (key == KEYSYMS.BACKSLASH) then gr = -1 end
		if (gr ~= nil) then
      if (gr == -1) then gr = nil end
      for _, uid in ipairs(GetSelectedUnits()) do  
        unit2group[GetUnitDefID(uid)] = gr
			end
		end
	end
end


function widget:GetConfigData() 
  local ret = {}
  for id, gr in pairs(unit2group) do 
    table.insert(ret, {UnitDefs[id].name, gr})
  end 
  return ret
end
	
function widget:SetConfigData(data) 
  if (data ~= nil) then
    for _, nam in ipairs(data) do
      local gr = UnitDefNames[nam[1]]
      if (gr ~= nil) then
        unit2group[gr.id] = nam[2]
      end
    end
  end
end

--------------------------------------------------------------------------------
