--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--
--  Copyright (C) 2007.
--  Licensed under the terms of the GNU GPL, v2 or later.
--
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function widget:GetInfo()
  return {
    name      = "CustomFormations",
    desc      = "Allows you to draw your own formation line.",
    author    = "jK",
    date      = "Sep 12, 2007",
    license   = "GNU GPL, v2 or later",
    layer     = 10000,
    enabled   = true,  --  loaded by default?
    handler   = true,
  }
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local gl = gl
local Spring = Spring

local formationVertices = {}
local totaldxy = 0  --// moved mouse distance 

local dimmVertices = {}
local alphaDimm = 1

local cmdTag = CMD.MOVE
local inButtonEvent = false  --//if you click the command menu the move/fight command is handled with left click instead of right one

local invertQueueKey = (Spring.GetConfigInt("InvertQueueKey", 0) == 1)

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function widget:MousePress(x, y, button)
  local _,activeid = Spring.GetActiveCommand()
  inButtonEvent = (activeid) and ((activeid==CMD.FIGHT) or (activeid==CMD.MOVE)) and (button==1)

  if (inButtonEvent) or ((activeid==nil)and(button==3)) then

    local selcnt = Spring.GetSelectedUnitsCount()
    if (selcnt>1) then

      local _,defid    = Spring.GetDefaultCommand()
      cmdTag = activeid or defid   --// CMD.MOVE or CMD.FIGHT

      if (defid==CMD.MOVE) or (inButtonEvent) then
        local units = Spring.GetSelectedUnits()
        units.n = nil

        --// check if there are moveable units selected
        local moveableUnits = false
        for _,unitID in ipairs(units) do
          local unitDefID = Spring.GetUnitDefID(unitID)
          if (UnitDefs[unitDefID].canMove) then moveableUnits = true; break; end
        end

        if (moveableUnits) then
          local _,pos = Spring.TraceScreenRay(x,y,true)

          if (pos) then
            widgetHandler:UpdateWidgetCallIn("DrawWorld", self)
            table.insert(formationVertices,pos)
            totaldxy = 0
            return true
          end

        end
      end
    end
  end

  return false
end


function widget:MouseMove(x, y, dx, dy, button)
  if ((inButtonEvent)and(button==3)) or
     ((not inButtonEvent)and(button~=3))
  then
    return false
  end

  if (#formationVertices>0) then
    totaldxy = totaldxy + dx^2 + dy^2
    if (totaldxy>40) then
      local _,pos = Spring.TraceScreenRay(x,y,true)
      if (pos) then
        table.insert(formationVertices,pos)
        totaldxy = 0
      end
      return true
    end
  end

  return false
end


function widget:MouseRelease(x, y, button)
  --// create modkeystate list   --// (keyState is for GiveOrder and keyState2 is for CommandNotify)
  local alt, ctrl, meta, shift = Spring.GetModKeyState()
  local keyState,keyState2 = {},{coded=0,alt=false,ctrl=false,shift=false,right=false}    
  if alt   then table.insert(keyState,"alt");  keyState2.alt =true;  end
  if ctrl  then table.insert(keyState,"ctrl"); keyState2.ctrl=true;  end
  if meta  then table.insert(keyState,"meta")                        end
  if (not inButtonEvent) then                  keyState2.right=true; end
  if (not inverseQueueKey and shift)or(inverseQueueKey and not shift) then table.insert(keyState,"shift"); keyState2.shift=true end

  --// calc the "coded" number of keyState2
  if keyState2.alt   then keyState2.coded = keyState2.coded + CMD.OPT_ALT   end
  if keyState2.ctl   then keyState2.coded = keyState2.coded + CMD.OPT_CTRL  end
  if keyState2.shift then keyState2.coded = keyState2.coded + CMD.OPT_SHIFT end
  if keyState2.right then keyState2.coded = keyState2.coded + CMD.OPT_RIGHT end

  --// end button event (user clicked command menu)
  if (inButtonEvent) then
    Spring.SetActiveCommand(-1)
  end

  --// single click? (no line drawn)
  if (#formationVertices==1) then
    Spring.GiveOrder(cmdTag, formationVertices[1], keyState)
    formationVertices = {}
    return -1
  end

  --// spread units out
  if (#formationVertices>0) then

    --// check if some widgets want to stop the command
    local retval = widgetHandler:CommandNotify(cmdTag, formationVertices[1], keyState2) or false
    if (retval) then
      formationLineEnds = {0}
      formationVertices = {}
      return -1
    end

    local units  = Spring.GetSelectedUnitsSorted()
    units.n = nil

    --// count moveable units
    local movunits = 0
    for unitDefID,unitIDs in pairs(units) do
      if (UnitDefs[unitDefID].canMove) then
        movunits = movunits + #unitIDs
      end
    end

    --// detect formation line length
    local formationLength = 0
    for i=2,#formationVertices do
      local lv = formationVertices[i-1]
      local v  = formationVertices[i]
      formationLength = formationLength + math.sqrt( (lv[1]-v[1])^2 + (lv[3]-v[3])^2 )
    end

    local unitdist = formationLength / (movunits-1)   --// movunits == 1 -> math.huge  , no segm ;)
    local idx,idxLength = 1,0

    --// send orders
    for unitDefID,unitIDs in pairs(units) do
      if (UnitDefs[unitDefID].canMove) then
        unitIDs.n = nil
        for _,unitID in ipairs(unitIDs) do
          local pos = formationVertices[idx]
          Spring.GiveOrderToUnit(unitID, cmdTag, pos, keyState)

          --// determine next unit order pos
          for i=idx+1,#formationVertices+1 do
            if (i>#formationVertices) then
              idx = #formationVertices
              break
            end
            local lv = formationVertices[i-1]
            local v  = formationVertices[i]
            local bfr_idxLength = idxLength + math.sqrt( (lv[1]-v[1])^2 + (lv[3]-v[3])^2 )
            if (unitdist<bfr_idxLength) then
              local fracLength = unitdist - idxLength
              local d = math.sqrt( (lv[1]-v[1])^2 + (lv[3]-v[3])^2 )
              local frac = fracLength/d

              --// add an in-between vertex
              table.insert(formationVertices,i,{lv[1]+frac*(v[1]-lv[1]),lv[2]+frac*(v[2]-lv[2]),lv[3]+frac*(v[3]-lv[3])})

              idxLength = 0
              idx = i
              break
            end
            if (unitdist==idxLength) then
              idxLength = 0
              idx = i
              break
            end
            idxLength = bfr_idxLength
          end

        end
      end
    end

    dimmVertices = formationVertices
    alphaDimm = 1
    formationVertices = {}
  end
  
  return -1
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------


function widget:DrawWorld()
  if (#formationVertices<1)and(#dimmVertices<1) then
    widgetHandler:RemoveWidgetCallIn("DrawWorld", self)
  end

  --// draw the lines
  gl.LineStipple(2, 4095)
  gl.LineWidth(2.0)

  if (cmdTag==CMD.MOVE)
    then gl.Color(0.5,1,0.5)
    else gl.Color(0.5,0.5,1) end

  gl.BeginEnd(GL.LINE_STRIP, function(vertices)
    for _,v in ipairs(vertices) do
      gl.Vertex(v[1],v[2],v[3])
    end
  end, formationVertices)

  if (#dimmVertices>1) then
    if (cmdTag==CMD.MOVE)
      then gl.Color(0.5,1,0.5,alphaDimm)
      else gl.Color(0.5,0.5,1,alphaDimm) end

    gl.BeginEnd(GL.LINE_STRIP, function(vertices)
      for _,v in ipairs(vertices) do
        gl.Vertex(v[1],v[2],v[3])
      end
    end, dimmVertices)
    alphaDimm = alphaDimm - 0.03
    if (alphaDimm<=0) then dimmVertices = {} end
  end

  gl.Color(1,1,1,1)
  gl.LineWidth(1.0)
  gl.LineStipple(false)
end